'''  Methods that check, repair, the local PASTA-ELN installation: no Qt-here '''
import json
import logging
import os
import platform
import shutil
import sys
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Callable, Optional
from .backend import Backend
from .fixedStringsJson import CONF_FILE_NAME, configurationGUI, defaultConfiguration
from .textTools.stringChanges import outputString


def getOS() -> str:
  '''
  Get operating system and python environment

  Returns:
    string: os + pythonEnvironment
  '''
  operatingSys = platform.system()
  # Get base/real prefix, or sys.prefix if there is none
  get_base_prefix_compat = getattr(sys, 'base_prefix', None) or getattr(sys, 'real_prefix', None) or sys.prefix
  in_virtualenv = get_base_prefix_compat != sys.prefix
  environment = sys.prefix if in_virtualenv else '_system_'
  return f'{operatingSys} {environment}'


def createDefaultConfiguration(pathPasta:Optional[Path]=None) -> dict[str,Any]:
  '''
  Create base of configuration file
  - basic project group
  - defaultProjectGroup
  - userID

  Args:
    pathPasta (Path): place to store pasta data

  Returns:
    dict: dictionary of configuration
  '''
  if pathPasta is None:
    if platform.system()=='Windows':
      pathPasta = Path.home()/'Documents'/'PASTA_ELN'
    else:
      pathPasta = Path.home()/'PASTA_ELN'
  addOnDir = Path(__file__).parent/'AddOns'
  conf: dict[str, Any] = {
      'defaultProjectGroup': 'research',
      'projectGroups': {
          'research': {
              'local': {'database': 'research', 'path': str(pathPasta)},
              'remote': {},
              'addOnDir': str(addOnDir),
              'addOns': {'project': {}, 'extractors':{}, 'table':{}}
          }},
      'version': 3}
  try:
    conf['userID']      = os.getlogin()
  except Exception:   #github action
    conf['userID']      = 'github_user'
  #create pastaDir if it does not exist
  if not pathPasta.is_dir():
    pathPasta.mkdir()
  return conf


def configuration(command:str, pathData:str) -> str:
  '''
  Check configuration file for consistencies

  Args:
    command (str): 'test' or 'repair'
    pathData (Path): path to use for data

  Returns:
    string: ''=success, else error messages
  '''
  logging.info('Configuration starting ...')
  output = ''
  if Path(pathData).is_dir():
    pathPasta = Path(pathData).absolute()
  else:
    pathPasta = Path.home()/pathData
    pathPasta.mkdir(exist_ok=True)
  try:
    with open(Path.home()/CONF_FILE_NAME, encoding='utf-8') as fConf:
      conf = json.load(fConf)
  except Exception:
    output += '**INFO configuration file does not exist\n'
    conf = createDefaultConfiguration(pathPasta) if command == 'repair' else {}
  logging.info(json.dumps(conf, indent=2))

  #check normal items
  for k,v in defaultConfiguration.items():
    if k not in conf:
      if command == 'repair':
        if isinstance(v,str) and v.startswith('$') and v.startswith('$'):
          v = eval(v[1:-1]) # pylint: disable=eval-used
        conf[k] = v
      else:
        output += outputString('text','error', f'No {k} in config file')
  #GUI items
  if 'GUI' not in conf:
    if command == 'repair':
      conf['GUI'] = {}
    else:
      output += outputString('text','error', 'No GUI in config file')
  for _, items in configurationGUI.items():
    for k,v in items.items():
      if 'GUI' in conf and k not in conf['GUI']:
        if command == 'repair':
          conf['GUI'][k] = v[1]
        else:
          output += outputString('text','error', f'No {k} in GUI part of config file')
  if command == 'repair':
    with open(Path.home()/CONF_FILE_NAME,'w', encoding='utf-8') as f:
      f.write(json.dumps(conf,indent=2))
  logging.info('Configuration ending')
  return output


def exampleData(force:bool=False, callbackPercent:Optional[Callable[[int],None]]=None, projectGroup:str='', outputFormat:str='print') -> str:
  '''
  Create example data after installation

  Args:
    force (bool): force creation by removing content before creation
    callbackPercent (function): callback function given to exampleData, such that exampleData can report progress back
    projectGroup (str): project group to create example data in
    outputFormat (str): output of the example data creation, see miscTools.outputString()
  '''
  logging.info('Start example data creation')
  if callbackPercent is not None:
    callbackPercent(0)
  if force:
    backend = Backend(projectGroup)
    dirName = backend.basePath
    backend.exit()
    try:
      shutil.rmtree(dirName)
      os.makedirs(dirName)
      if platform.system()=='Windows':
        print('Try-Except unnecessary')
    except Exception:
      pass
  if callbackPercent is not None:
    callbackPercent(1)
  backend = Backend(projectGroup)
  if callbackPercent is not None:
    callbackPercent(2)
  ### CREATE PROJECTS AND SHOW
  outputString(outputFormat,'h2','CREATE EXAMPLE PROJECT AND SHOW')
  backend.addData('x0', {'name': 'PASTAs Example Project', '.objective': 'Test if everything is working as intended.',
                         '.status': 'active', 'comment': 'Can be used as reference or deleted', 'tags':['Important']})
  if callbackPercent is not None:
    callbackPercent(3)
  outputString(outputFormat,'info', backend.output('x0'))
  if callbackPercent is not None:
    callbackPercent(4)
  logging.info('Finished creating example project')

  ### TEST PROJECT PLANING
  outputString(outputFormat,'h2','TEST PROJECT PLANING')
  dfProj  = backend.db.getView('viewDocType/x0')
  projID1 = list(dfProj['id'])[0]
  if callbackPercent is not None:
    callbackPercent(5)
  backend.changeHierarchy(projID1)
  if callbackPercent is not None:
    callbackPercent(6)
  backend.addData('x1',    {'comment': 'This is hard!', 'name': 'This is an example task', 'tags':['TODO']})
  if callbackPercent is not None:
    callbackPercent(7)
  currentID = backend.addData('x1', {'comment': 'This will take a long time.', 'tags':['WAIT'],
                                     'name': 'This is another example task'})['id']
  if callbackPercent is not None:
    callbackPercent(8)
  backend.changeHierarchy(currentID)
  if backend.cwd is not None:
    data2DirName = backend.basePath/backend.cwd
  else:
    return '**ERROR: backend is incorrect'
  backend.addData('x1',    {'name': 'This is an example subtask',     'comment': 'Random comment 1'})
  if callbackPercent is not None:
    callbackPercent(9)
  backend.addData('x1',    {'name': 'This is another example subtask','comment': 'Random comment 2'})
  if callbackPercent is not None:
    callbackPercent(10)
  backend.changeHierarchy(None)
  semStepID = backend.addData('x1',    {'name': 'Data files'})['id']
  if callbackPercent is not None:
    callbackPercent(11)
  backend.changeHierarchy(semStepID)
  if backend.cwd is not None:
    dataDirName = backend.basePath/backend.cwd
  else:
    return '**ERROR: backend is incorrect'
  backend.changeHierarchy(None)
  outputString(outputFormat,'info',backend.outputHierarchy(addID=True))
  if callbackPercent is not None:
    callbackPercent(12)
  logging.info('Finished project planning')

  ### TEST PROCEDURES
  outputString(outputFormat,'h2','TEST PROCEDURES')
  sopDir = backend.basePath/'StandardOperatingProcedures'
  os.makedirs(sopDir, exist_ok=True)
  with open(sopDir/'Example_SOP.md','w', encoding='utf-8') as fOut:
    fOut.write('# Put sample in instrument\n# Do something\nDo not forget to\n- not do anything wrong\n- **USE BOLD LETTERS**\n')
  if callbackPercent is not None:
    callbackPercent(13)
  backend.addData('procedure', {'name': 'StandardOperatingProcedures/Example_SOP.md', 'tags':['v1']})
  if callbackPercent is not None:
    callbackPercent(14)
  outputString(outputFormat,'info',backend.output('procedure'))
  df = backend.db.getView('viewDocType/procedure')
  procedureID = df[df['name']=='Example_SOP.md']['id'].values[0]
  if callbackPercent is not None:
    callbackPercent(15)
  logging.info('Finished procedures creating')

  ### TEST SAMPLES
  outputString(outputFormat,'h2','TEST SAMPLES')
  #TODO: qrCodes correct??
  backend.addData('sample',    {'name': 'Example sample', '.chemistry': 'A2B2C3', 'qrCodes': '13214124 99698708', 'comment':'this sample has multiple groups of metadata',
                                'geometry.height':4, 'geometry.width':2, 'weight.initial':6})
  if callbackPercent is not None:
    callbackPercent(16)
  outputString(outputFormat,'info',backend.output('sample'))
  if callbackPercent is not None:
    callbackPercent(17)
  outputString(outputFormat,'info',backend.outputQR())
  if callbackPercent is not None:
    callbackPercent(18)
  logging.info('Finished samples creating')

  ### ADD INSTRUMENTS AND THEIR ATTACHMENTS
  outputString(outputFormat,'h2','ADD INSTRUMENTS AND ATTACHMENTS')
  backend.addData('instrument', {'name': 'Big instrument', '.vendor':'Company A', '.model':'ABC-123', 'comment':'Instrument onto which attachments can be added'})
  backend.addData('instrument/extension', {'name': 'Sensor', '.vendor':'Company B', '.model':'org.comp.98765', 'comment':'Attachment that increases functionality of big instrument'})
  df = backend.db.getView('viewDocType/instrument')
  idInstrument = df[df['name']=='Big instrument']['id'].values[0]
  idSensor = df[df['name']=='Sensor']['id'].values[0]
  backend.db.initAttachment(idInstrument, 'Right side of instrument', 'instrument/extension')
  backend.db.addAttachment(idInstrument, 'Right side of instrument',
         {'date':datetime.now().isoformat(),'remark':'Worked well','docID':idSensor,'user':'nobody'})
  backend.db.addAttachment(idInstrument, 'Right side of instrument',
         {'date':(datetime.now()+timedelta(hours=1)).isoformat(),'remark':'Service','docID':'','user':'nobody'})
  outputString(outputFormat,'info',backend.output('instrument'))

  ###  TEST MEASUREMENTS AND SCANNING/CURATION
  outputString(outputFormat,'h2','TEST MEASUREMENTS AND SCANNING')
  shutil.copy(Path(__file__).parent/'Resources'/'ExampleMeasurements'/'simple.png', dataDirName)
  shutil.copy(Path(__file__).parent/'Resources'/'ExampleMeasurements'/'simple.csv', dataDirName)
  shutil.copy(Path(__file__).parent/'Resources'/'ExampleMeasurements'/'story.odt',  dataDirName)
  if callbackPercent is not None:
    callbackPercent(19)
  logging.info('Finished copy files')
  backend.scanProject(None, projID1)
  logging.info('Finished scan tree')
  if callbackPercent is not None:
    callbackPercent(20)

  ### USE GLOBAL FILES
  outputString(outputFormat,'h2','USE GLOBAL FILES')
  backend.changeHierarchy(semStepID)
  if callbackPercent is not None:
    callbackPercent(21)
  backend.addData('measurement', {
    'name'   :'https://upload.wikimedia.org/wikipedia/commons/thumb/a/a4/Misc_pollen.jpg/315px-Misc_pollen.jpg',\
    'comment':'- Remote image from wikipedia. Used for testing and reference\n- This item links to a procedure that was used for its creation.'
              '\n- One can link to samples, etc. to create complex metadata\n- This item also has a rating', 'tags':['_3'],
    '.procedure':procedureID })
  if callbackPercent is not None:
    callbackPercent(22)
  outputString(outputFormat,'info',backend.output('measurement'))
  if callbackPercent is not None:
    callbackPercent(23)
  logging.info('Finished global files additions')

  ###  TEST MEASUREMENTS AND SCANNING/CURATION 2
  outputString(outputFormat,'h2','TEST MEASUREMENTS AND SCANNING 2')
  shutil.copy(Path(__file__).parent/'Resources'/'ExampleMeasurements'/'simple.png', data2DirName)
  logging.info('Finished copy files 2')
  backend.scanProject(None, projID1)
  logging.info('Finished scan tree 2')
  df = backend.db.getView('viewDocType/measurement')
  docID = df[df['name']=='simple.png']['id'].values[0]
  doc   = backend.db.getDoc(docID)
  doc['comment'] = '# File with two locations\n - The same file can be located in different locations across different projects within one project group.'\
                   '\n - Since it is the same file, they share the same metadata: same comment, same tags, ...'\
                   '\n# These .png files use the data-science concept of schemata and ontology\n - The files have a agreed upon name and a custom convenience name (i.e. in german or french)'\
                   '\n - The also have a PID / PURL to an ontology node.\n - Units are also supported, obviously.'
  backend.editData(doc)
  docID = df[df['name']=='simple.csv']['id'].values[0]
  doc   = backend.db.getDoc(docID)
  doc['comment'] = '# These .csv files use the simple concept of units for metadata entries'
  backend.editData(doc)
  outputString(outputFormat,'info',backend.output('measurement'))

  ### VERIFY DATABASE INTEGRITY
  outputString(outputFormat,'h2','VERIFY DATABASE INTEGRITY')
  outputString(outputFormat,'info',backend.checkDB(outputStyle='text'))
  outputString(outputFormat,'h2','DONE WITH VERIFY')
  if callbackPercent is not None:
    callbackPercent(24)
  logging.info('Finished checking database')
  return 'Finished checking database'


def createShortcut() -> None:
  """
  Create alias and shortcut icon depending on operating system
  """
  logging.info('Create shortcut starting')
  if platform.system()=='Linux':
    content ='[Desktop Entry]\nName=PASTA ELN\nComment=PASTA electronic labnotebook\n'
    if sys.prefix==sys.base_prefix:   #normal installation into user-space
      content+='Exec=pastaELN\n'
    else:                             #installation in a virtual environment
      logging.info('In virtual environment, create an alias')
      with open(Path.home()/'.bashrc','a', encoding='utf-8') as fOut:
        alias = f"alias pastaELN='{sys.prefix}/bin/python3 -m pasta_eln.gui'"
        logging.info(alias)
        fOut.write(f'{alias}\n')
      content += f'Exec={sys.prefix}/bin/python3 -m pasta_eln.gui\n'
    content+='Icon='+ (Path(__file__).parent/'Resources'/'Icons'/'favicon64.png').as_posix() + '\n'
    content+='Terminal=false\nType=Application\nCategories=Science;Application;\n'
    try:
      linkString = (Path.home()/'Desktop'/'pastaELN.desktop').as_posix()
      with open(linkString,'w', encoding='utf-8') as fOut:
        fOut.write(content)
        os.system(f'gio set {linkString} metadata::trusted true') #for ubuntu systems
        os.chmod(Path.home()/'Desktop'/'pastaELN.desktop', 0o775)
    except Exception:
      pass
    try:
      with open(Path.home()/'.local'/'share'/'applications'/'pastaELN.desktop','w', encoding='utf-8') as fOut:
        fOut.write(content)
        os.chmod(Path.home()/'.local'/'share'/'applications'/'pastaELN.desktop', 0o777)
    except Exception:
      pass
  elif platform.system()=='Windows':
    import winshell
    from win32com.client import Dispatch
    shell = Dispatch('WScript.Shell')
    shortcut = shell.CreateShortCut( os.path.join(winshell.desktop(), 'pastaELN.lnk') )
    if env := os.environ.get('CONDA_PREFIX', ''):
      env = env.split('\\')[-1]  #just the env name
      user = os.getlogin()
      batContent = f"cmd.exe /c \"C:\\Users\\{user}\\anaconda3\\Scripts\\activate {env} && python -m pasta_eln.gui\""
      batLocation= f"C:\\Users\\{user}\\startPastaELN.bat"
      with open(batLocation, 'w', encoding='utf-8') as fBat:
        fBat.write(batContent)
      shortcut.Targetpath = batLocation
    else:
      shortcut.Targetpath = r'python -m pasta_eln.gui'
    shortcut.WorkingDirectory = str(Path.home())
    shortcut.IconLocation = str(Path(__file__).parent/'Resources'/'Icons'/'favicon64.ico')
    shortcut.save()
  logging.info('Create shortcut end')
  return


##############
# Main method for testing and installation without GUI
def main() -> None:
  ''' Main method and entry point for commands '''
  logPath = Path.home()/'pastaELN.log'
  #old versions of basicConfig do not know "encoding='utf-8'"
  logging.basicConfig(filename=logPath, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                      datefmt='%m-%d %H:%M:%S')   #This logging is always info, since for installation only
  for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
    logging.getLogger(package).setLevel(logging.WARNING)
  logging.info('Start PASTA Install')
  print('---- Test PASTA-ELN installation----')
  print('--   if nothing reported: it is ok.')
  print('getOS        :', getOS())
  res = configuration('test','.')
  flagConfiguration = 'ERROR' in res
  print(res)

  print('Add "install" argument to install PASTA-ELN.')
  if len(sys.argv)>2 and 'install' in sys.argv:
    print('---- Create PASTA-ELN installation ----')
    print(sys.argv)
    if flagConfiguration:
      print('repair  configuration:', configuration('repair', sys.argv[2] ))
  print('Add "example" argument to create example data.')
  if len(sys.argv)>1 and 'example' in sys.argv:
    print('---- Create Example data ----')
    print('create example data  :', exampleData())

  logging.info('End PASTA Install')
  return

# called by python3 -m pasta_eln.installTools
if __name__ == '__main__':
  main()



""" PastaConfigReaderFactory is a singleton class responsible for managing the configuration of the application. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: pasta_config_reader_factory.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import logging
from json import load
from os.path import exists, join
from pathlib import Path
from threading import Lock
from typing import Any
from PySide6 import QtCore
from PySide6.QtCore import QFileSystemWatcher
from pasta_eln.dataverse.config_error import ConfigError


class PastaConfigReaderFactory:
  """
  PastaConfigReaderFactory is a singleton class responsible for managing the configuration of the application.

  This class provides methods to read, update, and delete the configuration,
  as well as to monitor changes to the configuration file. It ensures that only one instance of the
  class exists and handles thread safety for configuration access.

  Attributes:
      _instance (Any): The singleton instance of the class.
      logger (Logger): Logger for logging configuration actions.
      config_file_name (str): The path to the configuration file.
      config (dict[str, Any]): The current configuration dictionary.
      mutex (Lock): A mutex for thread-safe access to the configuration.
      fs_watcher (QFileSystemWatcher): A file system watcher to monitor changes to the configuration file.

  Methods:
      get_instance: Retrieve the singleton instance of the class.
      __init__: Initialize the PastaConfigReaderFactory instance.
      read_pasta_config: Read the configuration file and update the internal configuration.
      config_file_changed: Handle changes to the configuration file.
      config: Property to get, set, and delete the current configuration.
  Note:
      This class is a singleton. It ensures that only one instance of the class exists.
      Make sure to use the get_instance() method to retrieve the singleton instance.
      Make sure that this class is initialized only after the instantiation of QApplication.instance() in-order to
        ensure that the file system watcher is properly set up and the signal-slot connections works as expected.
  """

  _instance: Any = None

  @classmethod
  def get_instance(cls) -> Any:
    """
    Retrieve the singleton instance of the class.

    This method ensures that only one instance of the class is created and returned. If an instance does not already exist, it creates a new one; otherwise, it returns the existing instance.

    Returns:
        Any: The singleton instance of the class PastaConfigReaderFactory.
    """
    if (not hasattr(cls, '_instance')
        or not getattr(cls, '_instance')):
      cls._instance = cls()
    return cls._instance

  def __init__(self) -> None:
    """
    Initialize the PastaConfigReaderFactory instance.

    This constructor sets up the logging, initializes the configuration file path, and prepares the file system
    watcher to monitor changes to the configuration file. It also reads the initial configuration and sets up a
    mutex for thread safety.
    """
    super().__init__()
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.config_file_name = join(Path.home(), '.pastaELN_v3.json')
    self.config: dict[str, Any] = {}
    self.mutex: Lock = Lock()
    self.fs_watcher = QFileSystemWatcher()
    self.fs_watcher.addPath(self.config_file_name)
    if exists(self.config_file_name):
      self.fs_watcher.fileChanged.connect(self.config_file_changed)
    self.read_pasta_config()

  def read_pasta_config(self) -> None:
    """
    Read the configuration file and update the internal configuration.

    This method logs the action of reading the configuration file and attempts to load its contents into the internal configuration dictionary.
    If the configuration file does not exist, it raises a ConfigError to indicate an issue with the installation.

    Raises:
        ConfigError: If the configuration file does not exist.
    """
    self.logger.info('Reading config file: %s', self.config_file_name)
    with self.mutex:
      if not exists(self.config_file_name):
        raise ConfigError('Config file not found, Corrupt installation!')
      with open(self.config_file_name, encoding='utf-8') as confFile:
        self.config = load(confFile)

  @QtCore.Slot()                                              # type: ignore[arg-type]
  def config_file_changed(self, config_file_path: str) -> None:
    """
    Handle changes to the configuration file.

    This method is triggered when the configuration file changes. It reads the updated configuration and refreshes the file system watcher to ensure it continues to monitor the correct file path.

    Args:
        config_file_path (str): The path to the configuration file that has changed.
    """

    self.read_pasta_config()
    if exists(config_file_path):
      self.fs_watcher.removePath(config_file_path)
      self.fs_watcher.addPath(config_file_path)

  @property
  def config(self) -> dict[str, Any] | None:
    """
    Get the current configuration.

    This property retrieves the current configuration stored in the instance. It ensures thread safety by using a
    mutex and returns the configuration dictionary if it exists; otherwise, it returns None.

    Returns:
        dict[str, Any] | None: The current configuration dictionary or None if not set.
    """

    with self.mutex:
      return self._config if hasattr(self, '_config') else None

  @config.setter
  def config(self, value: dict[str, Any] | None) -> None:
    """
    Set the configuration.

    This property allows the configuration to be updated with a new value. It accepts a dictionary representing the configuration or None to clear the existing configuration.

    Args:
        value (dict[str, Any] | None): The new configuration value to set.
    """

    self._config = value

  @config.deleter
  def config(self) -> None:
    """
    Delete the configuration.

    This property allows for the removal of the current configuration. It clears the stored configuration value, effectively resetting it to None.
    """

    del self._config



""" Http client for communicating with servers """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: webclient.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import logging
from typing import Any, Union
from aiohttp import BasicAuth, ClientResponse, ClientSession, ClientTimeout
from .utils import handle_http_client_exception


async def prepare_result(response: ClientResponse) -> dict[str, Any]:
  """
  Prepare the result from the http response
  Args:
    response (ClientResponse): The response object returned after GET/POST/DELETE/PUT Commands

  Returns:
    dict[str, Any]: The result is returned as a dictionary
      {
        "status": response.status,
        "headers": response.headers,
        "reason": response.reason,
        "result": response.result
      }

  """
  result = {
    'status': response.status,
    'headers': response.headers,
    'reason': response.reason
  }
  match response.headers.get('Content-Type'):
    case x if 'json' in x:  # type: ignore[operator]
      result['result'] = await response.json()
    case _:
      result['result'] = await response.text()
  return result


class AsyncHttpClient:
  """
  Asynchronous http client for communicating with REST based servers.
  """

  def __init__(self, session_timeout: int = 5) -> None:
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.session_timeout = session_timeout  # Timeout in seconds for the requests send to the lookup services
    self.session_request_errors: list[str] = []  # List of request errors

  @handle_http_client_exception
  async def get(self,
                base_url: str,
                request_params: Union[dict[str, Any], None] = None,
                request_headers: Union[dict[str, Any], None] = None,
                auth: Union[BasicAuth, None] = None,
                timeout: Union[int, None] = None) -> dict[str, Any]:
    """
    Send get request to the given url and parameters
    Args:
      base_url (str): Base url
      request_params (dict[str, Any]): Request parameters for the get request
      request_headers (dict[str, Any]): Request headers for the get request
      auth (BasicAuth): Basic authentication
      timeout (int): Timeout in seconds

    Returns (dict[str, Any]):
      The result is returned as a dictionary
      {
        "status": response.status,
        "headers": response.headers,
        "reason": response.reason,
        "result": response.result
      }
    """
    self.logger.info('Get url: %s', base_url)
    self.session_request_errors.clear()
    async with ClientSession() as session:
      async with session.get(base_url,
                             params=request_params,
                             headers=request_headers,
                             timeout=ClientTimeout(total=timeout if timeout is not None else self.session_timeout),
                             auth=auth) as response:
        result = await prepare_result(response)
    return result

  @handle_http_client_exception
  async def post(self,
                 base_url: str,
                 request_params: Union[dict[str, Any], None] = None,
                 request_headers: Union[dict[str, Any], None] = None,
                 json: Union[dict[str, Any], None] = None,
                 data: Union[Any, None] = None,
                 auth: Union[BasicAuth, None] = None,
                 timeout: Union[int, None] = None) -> dict[str, Any]:
    """
    Send post request to the given url and parameters
    Args:
      base_url (str): Base url
      request_params (dict[str, Any]): Request parameters for the post request
      request_headers (dict[str, Any]): Request headers for the post request
      json (dict[str, Any]): Json data for the post request
      data (Any): Data for the post request
      auth (BasicAuth): Basic authentication
      timeout (int): Timeout in seconds

    Returns:
      The result is returned as a dictionary
      {
        "status": response.status,
        "headers": response.headers,
        "reason": response.reason,
        "result": response.result
      }
    """
    self.logger.info('Post url: %s', base_url)
    self.session_request_errors.clear()
    async with ClientSession() as session:
      async with session.post(base_url,
                              headers=request_headers,
                              params=request_params,
                              timeout=ClientTimeout(total=timeout if timeout is not None else self.session_timeout),
                              json=json,
                              data=data,
                              auth=auth) as response:
        result = await prepare_result(response)
    return result

  @handle_http_client_exception
  async def delete(self,
                   base_url: str,
                   request_params: Union[dict[str, Any], None] = None,
                   request_headers: Union[dict[str, Any], None] = None,
                   json: Union[dict[str, Any], None] = None,
                   data: Union[Any, None] = None,
                   auth: Union[BasicAuth, None] = None,
                   timeout: Union[int, None] = None) -> dict[str, Any]:
    """
    Send delete request to the given url and parameters.
    Args:
      base_url (str): Base url
      request_params (dict[str, Any]): Request parameters for the delete request
      request_headers (dict[str, Any]): Request headers for the delete request
      json (dict[str, Any]): Json data for the delete request
      data (Any): Data for the delete request
      auth (BasicAuth): Basic authentication
      timeout (int): Timeout in seconds

    Returns:
      The result is returned as a dictionary
      {
        "status": response.status,
        "headers": response.headers,
        "reason": response.reason,
        "result": response.result
      }
    """
    self.logger.info('Delete url: %s', base_url)
    self.session_request_errors.clear()
    async with ClientSession() as session:
      async with session.delete(base_url,
                                headers=request_headers,
                                params=request_params,
                                timeout=ClientTimeout(total=timeout if timeout is not None else self.session_timeout),
                                json=json,
                                data=data,
                                auth=auth) as response:
        result = await prepare_result(response)
    return result



""" Color palette allows easy color access """
import platform
from PySide6.QtWidgets import QMainWindow
from qt_material import get_theme


class Palette():
  """ Color palette allows easy color access """
  def __init__(self, mainWindow:QMainWindow|None, theme:str) -> None:
    self.theme = theme
    if theme=='none':
      if platform.system() == 'Linux':
        self.subtheme = 'light' if mainWindow is None or mainWindow.palette().button().color().red()>128 else 'dark'
      else:
        self.subtheme = 'light'
      self.primary       = '#222222'
      self.secondaryText = '#000000'
    else:
      self.subtheme = 'light' if 'light' in theme else 'dark'
      self.primary       = get_theme(f'{theme}.xml')['primaryColor']
      self.secondaryText =  get_theme(f'{theme}.xml')['primaryTextColor']
    # for all themes
    if self.subtheme == 'dark':
      self.text       = '#EEEEEE'
      self.leafX      = '#222222'
      self.leafO      = '#333333'
      self.leafShadow = 'black'
    else:
      self.text       = '#111111'
      self.leafX      = '#EEEEEE'
      self.leafO      = '#FFFFFF'
      self.leafShadow = '#AAAAAA'


  def get(self, color:str, prefix:str) -> str:
    """ Get color with a prefix for CSS styling

    Args:
      color (str): qt-material colors without the trailing color; buttonText is an additional color
      prefix (str): css-key, e.g. 'background-color'

    Returns:
      str: css-string, e.g. 'background-color: #333421;'
    """
    if self.theme=='none':
      return ''
    if color=='buttonText':
      return  f'{prefix}: #EEEEEE; '
    return f'{prefix}: {get_theme(f"{self.theme}.xml")[f"{color}Color"]}; '



""" Data type class factory. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: data_type_class_factory.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
from typing import Any
from pasta_eln.GUI.dataverse.compound_data_type_class import CompoundDataTypeClass
from pasta_eln.GUI.dataverse.controlled_vocabulary_data_type_class import ControlledVocabularyDataTypeClass
from pasta_eln.GUI.dataverse.data_type_class_base import DataTypeClass
from pasta_eln.GUI.dataverse.data_type_class_context import DataTypeClassContext
from pasta_eln.GUI.dataverse.data_type_class_names import DataTypeClassName
from pasta_eln.GUI.dataverse.primitive_data_type_class import PrimitiveDataTypeClass


class DataTypeClassFactory:
  """
  Creates a new instance of the DataTypeClassFactory class.

  Explanation:
      This method creates a new instance of the DataTypeClassFactory class.

  Args:
      *_: Variable length argument list.
      **__: Arbitrary keyword arguments.

  Returns:
      Any: The new instance of the DataTypeClassFactory class.
  """

  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Creates a new instance of the DataTypeClassFactory class.

    Explanation:
        This method creates a new instance of the DataTypeClassFactory class.

    Args:
        *_: Variable length argument list.
        **__: Arbitrary keyword arguments.

    Returns:
        Any: The new instance of the DataTypeClassFactory class.

    """
    return super().__new__(cls)

  def __init__(self, context: DataTypeClassContext) -> None:
    """
    Initializes a new instance of the DataTypeClassFactory class.

    Explanation:
        This method initializes a new instance of the DataTypeClassFactory class.
    Args:
      context (DataTypeClassContext): The context of the data type class.
    """

    self.context: DataTypeClassContext = context
    self.factory_map: dict[DataTypeClassName, Any] = {
      DataTypeClassName.PRIMITIVE: self.make_primitive_data_type_class,
      DataTypeClassName.COMPOUND: self.make_compound_data_type_class,
      DataTypeClassName.CONTROLLED_VOCAB: self.make_controlled_vocab_data_type_class
    }

  def make_data_type_class(self, class_name: DataTypeClassName) -> DataTypeClass:
    """
    Creates an instance of a specific data type class based on the provided class name.

    Explanation:
        This method dynamically creates and returns an instance of a specific data type class based on the given class name.

    Args:
        class_name (DataTypeClassName): The name of the data type class to instantiate.

    Returns:
        DataTypeClass: An instance of the specified data type class.

    Raises:
        ValueError: If the provided data type class name is invalid.
    """
    if class_name in self.factory_map:
      return self.factory_map[class_name]()
    else:
      raise ValueError(f"Invalid data type class name: {class_name}")

  def make_primitive_data_type_class(self) -> DataTypeClass:
    """
    Creates an instance of a primitive data type class.

    Explanation:
        This method creates and returns an instance of a primitive data type class based on the context.

    Returns:
        DataTypeClass: An instance of a primitive data type class.
    """
    return PrimitiveDataTypeClass(self.context)

  def make_compound_data_type_class(self) -> DataTypeClass:
    """
    Creates an instance of a compound data type class.

    Explanation:
        This method creates and returns an instance of a compound data type class based on the context.

    Returns:
        DataTypeClass: An instance of a compound data type class.
    """
    return CompoundDataTypeClass(self.context)

  def make_controlled_vocab_data_type_class(self) -> DataTypeClass:
    """
    Creates an instance of a controlled vocabulary data type class.

    Explanation:
        This method creates and returns an instance of a controlled vocabulary data type class based on the context.

    Returns:
        DataTypeClass: An instance of a controlled vocabulary data type class.
    """
    return ControlledVocabularyDataTypeClass(self.context)



""" NameColumnDelegate module used for the table views """
from pandas import DataFrame
from PySide6.QtCore import QModelIndex, QPersistentModelIndex, Signal
from PySide6.QtWidgets import QLineEdit, QStyledItemDelegate, QStyleOptionViewItem, QWidget


class NameColumnDelegate(QStyledItemDelegate):
  """
  Delegate for creating the line edit with lookup icon for the iri column in data hierarchy editor tables
  """
  add_row_signal = Signal()

  def __init__(self, df:DataFrame, group:str) -> None:
    """
    Args:
      df (DataFrame): pandas dataframe containing the entire schema
      group (str): string of this group/class
    """
    super().__init__()
    self.df = df
    self.group = group


  def createEditor(self,
                   parent: QWidget,
                   option: QStyleOptionViewItem,
                   index:  QModelIndex | QPersistentModelIndex) -> QWidget:
    """
    Creates the line edit with lookup icon for the iri column cell represented by index

    Args:
      parent (QWidget): Parent table view.
      option (QStyleOptionViewItem): Style option for the cell represented by index.
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.

    Returns: QLineEdit widget
    """
    dfSub = self.df[self.df['class']==self.group]
    if index.row()>len(dfSub):
      return QWidget(parent)
    trues = dfSub[dfSub['idx']==index.row()]['name'].isin(['name','tags','comment']).values
    if len(trues)>0 and self.group=='' and trues[0]:
      return QWidget(parent)
    return QLineEdit(parent)


  def destroyEditor(self, editor:QWidget, index:QModelIndex | QPersistentModelIndex) -> None:
    """ destroy the line editor

    Args:
      editor (QWidget): Parent table view
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.
    """
    dfSub = self.df[self.df['class']==self.group]
    if index.row()==len(dfSub):
      self.add_row_signal.emit()
    super().destroyEditor(editor, index)
    return



""" Misc functions that do not require instances """
import importlib
import json
import logging
import os
import platform
import sys
import traceback
from collections.abc import Mapping
from io import BufferedReader
from pathlib import Path
from typing import Any, Union
from urllib import request
from .fixedStringsJson import CONF_FILE_NAME


class Bcolors:
  """
  Colors for Command-Line-Interface and output
  """
  if platform.system()=='Windows':
    HEADER = ''
    OKBLUE = ''
    OKGREEN = ''
    WARNING = ''
    FAIL = ''
    ENDC = ''
    BOLD = ''
    UNDERLINE = ''
  else:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def generic_hash(path:Path, forceFile:bool=False) -> str:
  """
  Hash an object based on its mode.

  inspired by:
  https://github.com/chris3torek/scripts/blob/master/githash.py

  Example implementation:
      result = generic_hash(sys.argv[1])
      print('%s: hash = %s' % (sys.argv[1], result))

  Args:
    path (Path): path
    forceFile (bool): force to get shasum of file and not of link (False for gitshasum)

  Returns:
    string: shasum

  Raises:
    ValueError: shasum of directory not supported
  """
  if str(path).startswith('http'):                      #Remote file:
    try:
      with request.urlopen(path.as_posix().replace(':/','://'), timeout=60) as site:
        meta = site.headers
        size = int(meta.get_all('Content-Length')[0])
        return blob_hash(site, size)
    except Exception:
      logging.error('Could not download content / hashing issue %s \n%s',path.as_posix().replace(':/','://'),traceback.format_exc())
      return ''
  if path.is_dir():
    raise ValueError(f'This seems to be a directory {path.as_posix()}')
  if forceFile and path.is_symlink():
    path = path.resolve()
  shasum = ''
  if path.is_symlink():    #if link, hash the link
    shasum = symlink_hash(path)
  elif path.is_file():  #Local file
    with open(path, 'rb') as stream:
      shasum = blob_hash(stream, path.stat().st_size)
  return shasum


def symlink_hash(path:Path) -> str:
  """
  Return (as hash instance) the hash of a symlink.
  Caller must use hexdigest() or digest() as needed on
  the result.

  Args:
    path (string): path to symlink

  Returns:
    string: shasum of link, aka short string
  """
  from hashlib import sha1
  hasher = sha1()
  data = os.readlink(path).encode('utf8', 'surrogateescape')
  hasher.update(b'blob {len(data)}\0')
  hasher.update(data)
  return hasher.hexdigest()


def blob_hash(stream:BufferedReader, size:int) -> str:
  """
  Return (as hash instance) the hash of a blob,
  as read from the given stream.

  Args:
    stream (string): content to be hashed
    size (int): size of the content

  Returns:
    string: shasum

  Raises:
    ValueError: size given is not the size of the stream
  """
  from hashlib import sha1
  hasher = sha1()
  hasher.update(f'blob {size}\0'.encode('ascii'))
  nRead = 0
  while True:
    data = stream.read(65536)     # read 64K at a time for storage requirements
    if data == b'':
      break
    nRead += len(data)
    hasher.update(data)
  if nRead != size:
    raise ValueError(f'{stream.name}: expected {size} bytes, found {nRead} bytes')
  return hasher.hexdigest()


def updateAddOnList(projectGroup:str='') -> dict[str, Any]:
  """
  Rules:
  - each data-type in its own try-except
  - inside try: raise ValueError exception on failure/None
  - except empty: pass
  - all descriptions in type have to be small letters
  - if want to force to skip top datatypes and use one at bottom: if doctype... -> exception

  Args:
    projectGroup (str): project group to scan

  Returns:
    dict: dict with all add-ons
  """
  with open(Path.home()/CONF_FILE_NAME, encoding='utf-8') as f:
    configuration = json.load(f)
  if not projectGroup:
    projectGroup = configuration['defaultProjectGroup']
  directory = Path(configuration['projectGroups'][projectGroup]['addOnDir'])
  sys.path.append(str(directory))  #allow add-ons
  # Extractors
  verboseDebug = False
  extractorsAll = {}
  projectAll    = {}
  for fileName in os.listdir(directory):
    if fileName.endswith('.py') and fileName.startswith('extractor_'):
      #start with file
      with open(directory/fileName, encoding='utf-8') as fIn:
        if verboseDebug: print('\n'+fileName)
        lines = fIn.readlines()
        extractorsThis = {}
        ifInFile, headerState, header = False, True, []
        for idx,line in enumerate(lines):
          line = line.rstrip()
          if idx>0 and '"""' in line:
            headerState = False
          if headerState:
            line = line.replace('"""','')
            header.append(line)
            continue
          if 'if' in line and '#:' in line:
            if verboseDebug: print('line', line)
            specialType = line.split('==')[1].split(':')[0].strip(" '"+'"')
            if verboseDebug: print('  special',specialType)
            extractorsThis[specialType] = line.split('#:')[1].strip()
            ifInFile = True
          elif 'else:' in line and '#:' in line:
            print('**ERROR there should not be an else in the code')
          elif 'return' in line and 'recipe' in line and not ifInFile:
            if verboseDebug: print('line', line)
            if line.count('recipe')==1:
              linePart = line.split('recipe')[1].strip()
              linePart = linePart.split(':')[1].split(',')[0].strip(" '"+'"')
            elif line.count('recipe')==2:
              possLines = [i.strip() for i in lines if ('recipe' in i and '=' in i and 'def' not in i)]
              if len(possLines)==1:
                linePart=possLines[0].split('=')[1].strip(" '"+'"')
              elif len(possLines)>1:
                print(f'**Warning: Could not decipher {fileName} Take shortest!')
                linePart=sorted(possLines)[0].split('=')[1].strip(" '"+'"')
              else:
                print(f'**ERROR: Could not decipher {fileName} File does not work!')
                linePart=''
            extractorsThis[linePart]='Default'
            if verboseDebug:
              print('  return', linePart)
        if verboseDebug:
          print('Extractors', extractorsThis)
        ending = fileName.split('_')[1].split('.')[0]
        extractorsAll[ending]=extractorsThis
                    #header not used for now
    if fileName.endswith('.py') and fileName.startswith('project_'):
      name        = fileName[:-3]
      module      = importlib.import_module(fileName[:-3])
      description = module.description
      projectAll[name] = description
  #update configuration file
  confProjectGroup = configuration['projectGroups'][projectGroup]['addOns']
  confProjectGroup['extractors'] = extractorsAll
  confProjectGroup['project']    = projectAll
  confProjectGroup['table']      = {}
  with open(Path.home()/CONF_FILE_NAME,'w', encoding='utf-8') as f:
    f.write(json.dumps(configuration, indent=2))
  return {'addon directory':directory} | extractorsAll | projectAll


def restart() -> None:
  """
  Complete restart: cold restart
  """
  try:
    os.execv('pastaELN',[''])  #installed version
  except Exception:
    os.execv(sys.executable, ['python3','-m','pasta_eln.gui']) #started for programming or debugging
  return


class DummyProgressBar():
  """ Class representing a progressbar that does not do anything
  """
  def setValue(self, value:int) -> int:
    """
    Set value

    Args:
      value (int): value to be set
    """
    return value
  def show(self) -> None:
    """ show progress bar """
    return
  def hide(self) -> None:
    """ hide progress bar """
    return


# adapted from flatten-dict https://github.com/ianlini/flatten-dict
# - reduce dependencies and only have python 3 code
# - add conversion of dict to list if applicable
def flatten(d:dict[Any,Any], keepPastaStruct:bool=False) -> dict[object, Any]:
  """Flatten `Mapping` object.

  Args:
    d : dict-like object
        The dict that will be flattened.
    keepPastaStruct : bool
        keep pasta elements from flattening

  Returns:
    flat_dict : dict
  """
  enumerate_types=(list,)
  flatten_types = (Mapping,) + enumerate_types
  flat_dict = {}

  def dot_reducer(k1:object, k2:object) -> Union[str, object]:
    """ Reducer function """
    return k2 if k1 is None else f"{k1}.{k2}"

  def _flatten(_d:Union[Mapping[Any, Any], list[Any]], depth:int, parent:object=None) -> bool:
    """ Recursive function """
    key_value_iterable = (enumerate(_d) if isinstance(_d, enumerate_types) else _d.items())
    has_item = False
    for key, value in key_value_iterable:
      has_item = True
      flat_key = dot_reducer(parent, key)
      if isinstance(value, flatten_types):
        # recursively build the result
        has_child = _flatten(value, depth=depth + 1, parent=flat_key)
        if has_child or not isinstance(value, ()): # ignore the key in this level because it already has child key or its value is empty
          continue
      # add an item to the result
      if flat_key in flat_dict:
        raise ValueError(f"duplicated key '{flat_key}'")
      flat_dict[flat_key] = value
    return has_item

  # start recursive calling
  backup = {'type':d.pop('type',''), 'branch':d.pop('branch',''), 'tags':d.pop('tags',''),
            'gui':d.pop('gui',''), 'qrCodes':d.pop('qrCodes',''), '_ids':d.pop('_ids','')} \
           if keepPastaStruct else {}
  _flatten(d, depth=1)
  return flat_dict | {k:v for k,v in backup.items() if v}


def hierarchy(d:dict[str,Any]) -> dict[str,Any]:
  """Reverse flattening of dict-like object

  Args:
    d : dict-like object
      The dict that will be reversed

  Returns
    normalDict : dict
  """
  def dot_splitter(flat_key:str) -> tuple[str, ...]:
    """ split using the . symbol """
    keys = tuple(flat_key.split('.'))
    return keys

  def nested_set_dict(d:dict[str,Any], keys:tuple[str, ...], value:Any) -> None:
    """Set a value to a sequence of nested keys

    Args:
      d : Mapping
      keys : Sequence[str]
      value : Any
    """
    key = keys[0]
    if len(keys) == 1:
      if key in d:
        raise ValueError(f"duplicated key '{key}'")
      d[key] = value
      return
    d = d.setdefault(key, {})
    nested_set_dict(d, keys[1:], value)
    return

  def dict2list(d:dict[str,Any]) -> Union[list[dict[str,Any]], dict[str,Any]]:
    """ convert a dictionary to list if all keys are numbers """
    for key in d:
      if isinstance(d[key], dict):
        d[key] = dict2list(d[key])
    if all(i.isdigit() for i in d):
      d = list(d.values())                                                           # type: ignore[assignment]
    return d

  # start recursion
  normalDict:dict[str,Any] = {}
  for flat_key, value in d.items():
    key_tuple = dot_splitter(flat_key)
    nested_set_dict(normalDict, key_tuple, value)
  normalDict =  dict2list(normalDict)                                                # type: ignore[assignment]
  return normalDict



from PySide6.QtWidgets import QMainWindow
from pasta_eln.backend import Backend
from pasta_eln.guiCommunicate import Communicate
from pasta_eln.GUI.palette import Palette
from pasta_eln.GUI.config import Configuration

def test_simple(qtbot):
  app = QMainWindow()
  backend = Backend('research')
  palette = Palette(app,'none')
  comm = Communicate(backend,palette)
  dialog = Configuration(comm)
  dialog.setMinimumSize(1024,800)
  dialog.show()
  qtbot.addWidget(dialog)
  path = qtbot.screenshot(dialog)
  print(path)
